@@ -1,5 +1,6 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Antonio Diaz Diaz.
+    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012,
+    2013 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,6 +27,7 @@
 #include <ctime>
 #include <string>
 #include <vector>
+#include <stdint.h>
 #include <unistd.h>
 
 #include "block.h"
@@ -34,7 +36,17 @@
 
 namespace {
 
-int my_fgetc( FILE * f ) throw()
+void input_pos_error( const long long pos, const long long isize )
+  {
+  char buf[128];
+  snprintf( buf, sizeof buf,
+            "Can't start reading at pos %lld.\n"
+            "          Input file is only %lld bytes long.", pos, isize );
+  show_error( buf );
+  }
+
+
+int my_fgetc( FILE * const f )
   {
   int ch;
   bool comment = false;
@@ -49,13 +61,16 @@ int my_fgetc( FILE * f ) throw()
   }
 
 
-const char * my_fgets( FILE * f, int & linenum ) throw()
+// Read a line discarding comments, leading whitespace and blank lines.
+// Returns 0 if at EOF.
+//
+const char * my_fgets( FILE * const f, int & linenum )
   {
   const int maxlen = 127;
   static char buf[maxlen+1];
   int ch, len = 1;
 
-  while( len == 1 )
+  while( len == 1 )			// while line is blank
     {
     do { ch = my_fgetc( f ); if( ch == '\n' ) ++linenum; }
     while( std::isspace( ch ) );
@@ -78,7 +93,7 @@ void extend_sblock_vector( std::vector< Sblock > & sblock_vector,
   {
   if( sblock_vector.size() == 0 )
     {
-    Sblock sb( 0, (isize > 0) ? isize : -1, Sblock::non_tried );
+    Sblock sb( 0, ( isize > 0 ) ? isize : -1, Sblock::non_tried );
     sb.fix_size();
     sblock_vector.push_back( sb );
     return;
@@ -92,12 +107,22 @@ void extend_sblock_vector( std::vector< Sblock > & sblock_vector,
     {
     if( back.pos() >= isize )
       {
-      if( back.pos() == isize && back.status() == Sblock::non_tried )
+      if( back.pos() == isize && back.status() != Sblock::finished )
         { sblock_vector.pop_back(); return; }
-      show_error( "bad logfile; last block begins past end of input file" );
+      show_error( "Last block in logfile begins past end of input file.\n"
+                  "          Use '-C' if you are reading from a partial copy.",
+                  0, true );
+      std::exit( 1 );
+      }
+    if( end > isize )
+      {
+      if( back.status() != Sblock::finished )
+        { back.size( isize - back.pos() ); return; }
+      show_error( "Rescued data in logfile goes past end of input file.\n"
+                  "          Use '-C' if you are reading from a partial copy.",
+                  0, true );
       std::exit( 1 );
       }
-    if( end < 0 || end > isize ) back.size( isize - back.pos() );
     else if( end < isize )
       sblock_vector.push_back( Sblock( end, isize - end, Sblock::non_tried ) );
     }
@@ -110,23 +135,26 @@ void extend_sblock_vector( std::vector< Sblock > & sblock_vector,
   }
 
 
-void show_logfile_error( const char * filename, const int linenum )
+void show_logfile_error( const char * const filename, const int linenum )
   {
   char buf[80];
-  snprintf( buf, sizeof buf, "error in logfile %s, line %d", filename, linenum );
+  snprintf( buf, sizeof buf, "error in logfile %s, line %d.", filename, linenum );
   show_error( buf );
   }
 
 
-bool read_logfile( const char * name, std::vector< Sblock > & sblock_vector,
+// Returns true if logfile exists and is readable.
+//
+bool read_logfile( const char * const logname,
+                   std::vector< Sblock > & sblock_vector,
                    long long & current_pos, Logbook::Status & current_status )
   {
-  FILE *f = std::fopen( name, "r" );
+  FILE * const f = std::fopen( logname, "r" );
   if( !f ) return false;
   int linenum = 0;
   sblock_vector.clear();
 
-  const char *line = my_fgets( f, linenum );
+  const char * line = my_fgets( f, linenum );
   if( line )						// status line
     {
     char ch;
@@ -135,7 +163,7 @@ bool read_logfile( const char * name, std::vector< Sblock > & sblock_vector,
       current_status = Logbook::Status( ch );
     else
       {
-      show_logfile_error( name, linenum );
+      show_logfile_error( logname, linenum );
       show_error( "Are you using a logfile from ddrescue 1.5 or older?" );
       std::exit( 2 );
       }
@@ -149,14 +177,14 @@ bool read_logfile( const char * name, std::vector< Sblock > & sblock_vector,
       if( n == 3 && pos >= 0 && Sblock::isstatus( ch ) &&
           ( size > 0 || size == -1 || ( size == 0 && pos == 0 ) ) )
         {
-        Sblock::Status st = Sblock::Status( ch );
+        const Sblock::Status st = Sblock::Status( ch );
         Sblock sb( pos, size, st ); sb.fix_size();
         if( sblock_vector.size() > 0 && !sb.follows( sblock_vector.back() ) )
-          { show_logfile_error( name, linenum ); std::exit( 2 ); }
+          { show_logfile_error( logname, linenum ); std::exit( 2 ); }
         sblock_vector.push_back( sb );
         }
       else
-        { show_logfile_error( name, linenum ); std::exit( 2 ); }
+        { show_logfile_error( logname, linenum ); std::exit( 2 ); }
       }
     }
   std::fclose( f );
@@ -166,33 +194,35 @@ bool read_logfile( const char * name, std::vector< Sblock > & sblock_vector,
 } // end namespace
 
 
-Domain::Domain( const char * name, const long long p, const long long s )
+Domain::Domain( const long long p, const long long s,
+                const char * const logname )
   {
   Block b( p, s ); b.fix_size();
-  if( !name ) { block_vector.push_back( b ); return; }
+  if( !logname || !logname[0] ) { block_vector.push_back( b ); return; }
   std::vector< Sblock > sblock_vector;
   long long current_pos;			// not used
   Logbook::Status current_status;		// not used
-  if( !read_logfile( name, sblock_vector, current_pos, current_status ) )
+  if( !read_logfile( logname, sblock_vector, current_pos, current_status ) )
     {
     char buf[80];
     snprintf( buf, sizeof buf,
-              "logfile `%s' does not exist or is not readable", name );
+              "Logfile '%s' does not exist or is not readable.", logname );
     show_error( buf );
     std::exit( 1 );
     }
-  for( unsigned int i = 0; i < sblock_vector.size(); ++i )
+  for( unsigned i = 0; i < sblock_vector.size(); ++i )
     {
     const Sblock & sb = sblock_vector[i];
     if( sb.status() == Sblock::finished ) block_vector.push_back( sb );
     }
-  this->crop( b );
+  if( block_vector.size() == 0 ) block_vector.push_back( Block( 0, 0 ) );
+  else this->crop( b );
   }
 
 
 void Logbook::split_domain_border_sblocks()
   {
-  for( unsigned int i = 0; i < sblock_vector.size(); ++i )
+  for( unsigned i = 0; i < sblock_vector.size(); ++i )
     {
     Sblock & sb = sblock_vector[i];
     const long long pos = domain_.breaks_block_by( sb );
@@ -206,29 +236,39 @@ void Logbook::split_domain_border_sblocks()
   }
 
 
-Logbook::Logbook( const long long ipos, const long long opos, Domain & dom,
-                  const long long isize,
-                  const char * name, const int cluster, const int hardbs,
-                  const bool complete_only )
-  : offset_( opos - ipos ), current_pos_( 0 ), current_status_( copying ),
-    domain_( dom ), filename_( name ),
+Logbook::Logbook( const long long offset, const long long isize,
+                  Domain & dom, const char * const logname,
+                  const int cluster, const int hardbs,
+                  const bool complete_only, const bool do_not_read )
+  : offset_( offset ), current_pos_( 0 ), logfile_isize_( 0 ),
+    current_status_( copying ), domain_( dom ), filename_( logname ),
     hardbs_( hardbs ), softbs_( cluster * hardbs ),
     final_msg_( 0 ), final_errno_( 0 ), index_( 0 ),
-    ul_t1( std::time( 0 ) )
+    ul_t1( std::time( 0 ) ), logfile_exists_( false )
   {
   int alignment = sysconf( _SC_PAGESIZE );
   if( alignment < hardbs_ || alignment % hardbs_ ) alignment = hardbs_;
   if( alignment < 2 || alignment > 65536 ) alignment = 0;
-  iobuf_ = iobuf_base = new char[ softbs_ + alignment ];
+  iobuf_ = iobuf_base = new uint8_t[ softbs_ + alignment ];
   if( alignment > 1 )		// align iobuf for use with raw devices
     {
     const int disp = alignment - ( reinterpret_cast<long> (iobuf_) % alignment );
     if( disp > 0 && disp < alignment ) iobuf_ += disp;
     }
 
-  if( !domain_.crop_by_file_size( isize ) ) std::exit( 1 );
-  if( filename_ )
-    read_logfile( filename_, sblock_vector, current_pos_, current_status_ );
+  if( isize > 0 )
+    {
+    if( domain_.pos() >= isize )
+      { input_pos_error( domain_.pos(), isize ); std::exit( 1 ); }
+    domain_.crop_by_file_size( isize );
+    }
+  if( filename_ && !do_not_read )
+    {
+    logfile_exists_ =
+      read_logfile( filename_, sblock_vector, current_pos_, current_status_ );
+    if( logfile_exists_ && sblock_vector.size() )
+      logfile_isize_ = sblock_vector.back().end();
+    }
   if( !complete_only ) extend_sblock_vector( sblock_vector, isize );
   else if( sblock_vector.size() )  // limit domain to blocks read from logfile
     {
@@ -242,9 +282,9 @@ Logbook::Logbook( const long long ipos, const long long opos, Domain & dom,
   }
 
 
-bool Logbook::blank() const throw()
+bool Logbook::blank() const
   {
-  for( unsigned int i = 0; i < sblock_vector.size(); ++i )
+  for( unsigned i = 0; i < sblock_vector.size(); ++i )
     if( sblock_vector[i].status() != Sblock::non_tried )
       return false;
   return true;
@@ -253,85 +293,108 @@ bool Logbook::blank() const throw()
 
 void Logbook::compact_sblock_vector()
   {
-  for( unsigned int i = sblock_vector.size(); i >= 2; --i )
-    if( sblock_vector[i-2].join( sblock_vector[i-1] ) )
-      sblock_vector.erase( sblock_vector.begin() + i - 1 );
+  for( unsigned i = sblock_vector.size(); i >= 2; )
+    {
+    --i;
+    if( sblock_vector[i-1].join( sblock_vector[i] ) )
+      sblock_vector.erase( sblock_vector.begin() + i );
+    }
   }
 
 
 // Writes periodically the logfile to disc.
 // Returns false only if update is attempted and fails.
 //
-bool Logbook::update_logfile( const int odes, const bool force )
+bool Logbook::update_logfile( const int odes, const bool force,
+                              const bool retry )
   {
   if( !filename_ ) return true;
-  const int interval = 30 + std::min( 270, sblocks() / 40 );
+  const int interval = 30 + std::min( 270, sblocks() / 38 );	// 30s to 5m
   const long t2 = std::time( 0 );
   if( !force && t2 - ul_t1 < interval ) return true;
   ul_t1 = t2;
   if( odes >= 0 ) fsync( odes );
 
   errno = 0;
-  FILE *f = std::fopen( filename_, "w" );
+  FILE * const f = std::fopen( filename_, "w" );
   if( f )
     {
-    write_logfile_header( f );
-    std::fprintf( f, "# current_pos  current_status\n" );
-    std::fprintf( f, "0x%08llX     %c\n", current_pos_, current_status_ );
-    std::fprintf( f, "#      pos        size  status\n" );
-    for( unsigned int i = 0; i < sblock_vector.size(); ++i )
-      {
-      const Sblock & sb = sblock_vector[i];
-      std::fprintf( f, "0x%08llX  0x%08llX  %c\n", sb.pos(), sb.size(), sb.status() );
-      }
+    write_logfile( f );
     if( std::fclose( f ) == 0 ) return true;
     }
 
   if( verbosity >= 0 )
     {
     char buf[80];
-    const char * const s =
-      f ? "error writing logfile %s" : "error opening logfile %s for writing";
-    snprintf( buf, sizeof buf, s, filename_ );
-    std::fprintf( stderr, "\n" );
+    snprintf( buf, sizeof buf, f ? "Error writing logfile '%s'." :
+                                   "Error opening logfile '%s' for writing.",
+              filename_ );
+    if( retry ) std::fprintf( stderr, "\n" );
     show_error( buf, errno );
-    std::fprintf( stderr, "Fix the problem and press ENTER to retry, or Q+ENTER to abort. " );
-    std::fflush( stderr );
-    while( true )
+    if( retry )
       {
-      const char c = std::tolower( std::fgetc( stdin ) );
-      if( c == '\r' || c == '\n' )
+      std::fprintf( stderr, "Fix the problem and press ENTER to retry, or Q+ENTER to abort. " );
+      std::fflush( stderr );
+      while( true )
         {
-        std::fprintf( stderr, "\n\n\n\n" );
-        return update_logfile( -1, true );
+        const char c = std::tolower( std::fgetc( stdin ) );
+        if( c == '\r' || c == '\n' )
+          {
+          std::fprintf( stderr, "\n\n\n\n" );
+          return update_logfile( -1, true );
+          }
+        if( c == 'q' ) break;
         }
-      if( c == 'q' ) break;
       }
-    std::fprintf( stderr, "\n\n\n\n" );
     }
   return false;
   }
 
 
-void Logbook::truncate_vector( const long long pos )
+void Logbook::write_logfile( FILE * const f ) const
   {
-  int i = sblocks() - 1;
-  while( i >= 0 && sblock_vector[i].pos() >= pos ) --i;
-  if( i < 0 )
+  write_logfile_header( f );
+  std::fprintf( f, "# current_pos  current_status\n" );
+  std::fprintf( f, "0x%08llX     %c\n", current_pos_, current_status_ );
+  std::fprintf( f, "#      pos        size  status\n" );
+  for( unsigned i = 0; i < sblock_vector.size(); ++i )
     {
-    sblock_vector.clear();
-    sblock_vector.push_back( Sblock( pos, 0, Sblock::non_tried ) );
-    }
-  else
-    {
-    Sblock & sb = sblock_vector[i];
-    if( sb.includes( pos ) ) sb.size( pos - sb.pos() );
-    sblock_vector.erase( sblock_vector.begin() + i + 1, sblock_vector.end() );
+    const Sblock & sb = sblock_vector[i];
+    std::fprintf( f, "0x%08llX  0x%08llX  %c\n", sb.pos(), sb.size(), sb.status() );
     }
   }
 
 
-int Logbook::find_index( const long long pos ) const throw()
+// Returns false only if truncation would remove finished blocks and
+// force is false.
+//
+bool Logbook::truncate_vector( const long long end, const bool force )
+  {
+  int i = sblocks() - 1;
+  while( i >= 0 && sblock_vector[i].pos() >= end ) --i;
+  if( i < 0 )
+    {
+    sblock_vector.clear();
+    sblock_vector.push_back( Sblock( 0, 0, Sblock::non_tried ) );
+    }
+  else
+    {
+    if( !force )
+      for( unsigned j = i + 1; j < sblock_vector.size(); ++j )
+        if( sblock_vector[j].status() == Sblock::finished ) return false;
+    Sblock & sb = sblock_vector[i];
+    if( sb.includes( end ) )
+      {
+      if( !force && sb.status() == Sblock::finished ) return false;
+      sb.size( end - sb.pos() );
+      }
+    sblock_vector.erase( sblock_vector.begin() + i + 1, sblock_vector.end() );
+    }
+  return true;
+  }
+
+
+int Logbook::find_index( const long long pos ) const
   {
   if( index_ < 0 || index_ >= sblocks() ) index_ = sblocks() / 2;
   while( index_ + 1 < sblocks() && pos >= sblock_vector[index_].end() )
@@ -343,10 +406,41 @@ int Logbook::find_index( const long long pos ) const throw()
   }
 
 
+int Logbook::find_largest_sblock( const Sblock::Status st ) const
+  {
+  long long size = 0;
+  int index = -1;
+  for( int i = 0; i < sblocks(); ++i )
+    {
+    const Sblock & sb = sblock_vector[i];
+    if( sb.status() == st && sb.size() > size && domain_.includes( sb ) )
+      { size = sb.size(); index = i; }
+    }
+  return index;
+  }
+
+
+int Logbook::find_smallest_sblock( const Sblock::Status st ) const
+  {
+  long long size = LLONG_MAX;
+  int index = -1;
+  for( int i = 0; i < sblocks(); ++i )
+    {
+    const Sblock & sb = sblock_vector[i];
+    if( sb.status() == st &&
+        ( sb.size() < size || ( sb.size() == size && index < 0 ) ) &&
+        domain_.includes( sb ) )
+      { size = sb.size(); index = i; if( size <= hardbs_ ) break; }
+    }
+  return index;
+  }
+
+
 // Find chunk from b.pos of size <= b.size and status st.
 // if not found, puts b.size to 0.
 //
-void Logbook::find_chunk( Block & b, const Sblock::Status st ) const
+void Logbook::find_chunk( Block & b, const Sblock::Status st,
+                          const int alignment ) const
   {
   if( b.size() <= 0 ) return;
   if( b.pos() < sblock_vector.front().pos() )
@@ -364,14 +458,15 @@ void Logbook::find_chunk( Block & b, const Sblock::Status st ) const
   if( !sblock_vector[index_].includes( b ) )
     b.crop( sblock_vector[index_] );
   if( b.end() != sblock_vector[index_].end() )
-    b.align_end( hardbs_ );
+    b.align_end( alignment ? alignment : hardbs_ );
   }
 
 
 // Find chunk from b.end backwards of size <= b.size and status st.
 // if not found, puts b.size to 0.
 //
-void Logbook::rfind_chunk( Block & b, const Sblock::Status st ) const
+void Logbook::rfind_chunk( Block & b, const Sblock::Status st,
+                           const int alignment ) const
   {
   if( b.size() <= 0 ) return;
   b.fix_size();
@@ -388,19 +483,39 @@ void Logbook::rfind_chunk( Block & b, const Sblock::Status st ) const
   if( !sblock_vector[index_].includes( b ) )
     b.crop( sblock_vector[index_] );
   if( b.pos() != sblock_vector[index_].pos() )
-    b.align_pos( hardbs_ );
+    b.align_pos( alignment ? alignment : hardbs_ );
   }
 
 
-void Logbook::change_chunk_status( const Block & b, const Sblock::Status st )
+// Returns an adjust value (-1, 0, +1) to keep "errors" updated without
+// having to call count_errors every time.
+//   - - -   -->   - + -   return +1
+//   - - +   -->   - + +   return  0
+//   - + -   -->   - - -   return -1
+//   - + +   -->   - - +   return  0
+//   + - -   -->   + + -   return  0
+//   + - +   -->   + + +   return -1
+//   + + -   -->   + - -   return  0
+//   + + +   -->   + - +   return +1
+//
+int Logbook::change_chunk_status( const Block & b, const Sblock::Status st )
   {
-  if( b.size() <= 0 ) return;
+  if( b.size() <= 0 ) return 0;
   if( !domain_.includes( b ) || find_index( b.pos() ) < 0 ||
       !domain_.includes( sblock_vector[index_] ) )
     internal_error( "can't change status of chunk not in rescue domain" );
   if( !sblock_vector[index_].includes( b ) )
     internal_error( "can't change status of chunk spread over more than 1 block" );
-  if( sblock_vector[index_].status() == st ) return;
+  if( sblock_vector[index_].status() == st ) return 0;
+
+  const bool old_st_good = Sblock::is_good_status( sblock_vector[index_].status() );
+  const bool new_st_good = Sblock::is_good_status( st );
+  bool bl_st_good = ( index_ <= 0 ||
+                      !domain_.includes( sblock_vector[index_-1] ) ||
+                      Sblock::is_good_status( sblock_vector[index_-1].status() ) );
+  bool br_st_good = ( index_ + 1 >= sblocks() ||
+                      !domain_.includes( sblock_vector[index_+1] ) ||
+                      Sblock::is_good_status( sblock_vector[index_+1].status() ) );
   if( sblock_vector[index_].pos() < b.pos() )
     {
     if( sblock_vector[index_].end() == b.end() &&
@@ -410,15 +525,17 @@ void Logbook::change_chunk_status( const Block & b, const Sblock::Status st )
       sblock_vector[index_].inc_size( -b.size() );
       sblock_vector[index_+1].pos( b.pos() );
       sblock_vector[index_+1].inc_size( b.size() );
-      return;
+      return 0;
       }
     insert_sblock( index_, sblock_vector[index_].split( b.pos() ) );
     ++index_;
+    bl_st_good = old_st_good;
     }
   if( sblock_vector[index_].size() > b.size() )
     {
     sblock_vector[index_].pos( b.end() );
     sblock_vector[index_].inc_size( -b.size() );
+    br_st_good = Sblock::is_good_status( sblock_vector[index_].status() );
     if( index_ > 0 && sblock_vector[index_-1].status() == st &&
         domain_.includes( sblock_vector[index_-1] ) )
       sblock_vector[index_-1].inc_size( b.size() );
@@ -441,4 +558,24 @@ void Logbook::change_chunk_status( const Block & b, const Sblock::Status st )
       erase_sblock( index_ + 1 );
       }
     }
+  int retval = 0;
+  if( new_st_good != old_st_good && bl_st_good == br_st_good )
+    { if( old_st_good == bl_st_good ) retval = +1; else retval = -1; }
+  return retval;
+  }
+
+
+const char * Logbook::status_name( const Logbook::Status st )
+  {
+  switch( st )
+    {
+    case copying:    return "copying";
+    case trimming:   return "trimming";
+    case splitting:  return "splitting";
+    case retrying:   return "retrying";
+    case filling:    return "filling";
+    case generating: return "generating";
+    case finished:   return "finished";
+    }
+  return "unknown";			// should not be reached
   }
